#!/usr/bin/env python3
r"""
Used by CI to install a new cluster and run tests.

Env vars needed:
- COSMOS_API_TOKEN

Env vars optional:
- NO_TERMINATE_DROPLETS: list of droplet ids that should not be
  removed when deleting droplets.

Install requirements:

- On Alpine, install with:

    apk --no-cache add python3-dev build-base libffi-dev linux-headers \
      openssl-dev openssh-client
    pip3 install -r requirements.txt
"""

import argparse
import logging
import os
import sys
from math import floor
from subprocess import PIPE, Popen

import greenhost_cloud

from stackspin import ansible, cluster, name

# We're limiting to 50, because we use subdomains, the current longest of which
# is 7 characters (`office.`). Max CN length is 63 characters, so we have some
# wiggle room for longer subdomains in the future.
# See https://github.com/jetstack/cert-manager/issues/2794 for details.
MAX_DOMAIN_LENGTH = 50


def main():  # pylint: disable=too-many-statements,too-many-branches,too-many-locals
    """
    Parse arguments and start program. Depending on the arguments parsed,
    one of three functions is called:

    - create, responsible for setting up VPSs and their local configuration
      files
    - install, responsible for setting up Kubernetes and Stackspin on those
      VPSs
    - test, responsible for testing if all the setup steps worked.

    Run python3 -m stackspin --help for more information.
    """
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        prog=name,
        description='Run installation script '
        'to deploy Stackspin to a given node. Note that you always need to '
        'provide a cluster name. For subcommand specific help, for example, '
        'run "python3 -m stackspin your-cluster install --help"')

    parser.add_argument(
        'cluster_name',
        metavar='CLUSTER_NAME',
        type=str,
        help='Name of the cluster you want to use stackspin with')

    subparsers = parser.add_subparsers(help='Available subcommands')

    create_parser = subparsers.add_parser('create', help='Creates a new cluster')
    create_parser.set_defaults(func=create)

    create_parser.add_argument(
        'domain',
        metavar='DOMAIN_NAME',
        help='Domain name to run Stackspin under')
    create_parser.add_argument(
        '--create-hostname',
        type=str,
        help='hostname of the machine. If not provided for a new machine, the '
             'cluster name is used.')

    create_parser.add_argument(
        '--docker-mirror-server',
        help=("Server name for a docker mirror"))
    create_parser.add_argument(
        '--docker-mirror-endpoint',
        help=("Full endpoint for a docker mirror"))
    create_parser.add_argument(
        '--docker-mirror-username',
        help=("Username for `docker login` to docker mirror"))
    create_parser.add_argument(
        '--docker-mirror-password',
        help=("Password for `docker login` to docker mirror"))


    group = create_parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        '--create-droplet',
        action='store_true',
        help='Create droplet automatically')
    group.add_argument(
        '--droplet-id',
        metavar='ID',
        type=int,
        help='ID of existing Greenhost VPS to use')
    group.add_argument(
        '--droplet-hostname',
        metavar='HOSTNAME',
        help='Hostname of existing Greenhost VPS to use')
    group.add_argument(
        '--ip-address',
        metavar='IP',
        help='IP address of non-greenhost VPS to use. You *must* provide '
             '--create-hostname with this argument')

    droplet_creation_group = create_parser.add_argument_group(
        'droplet creation',
        'When using --create-droplet, you need to provide:')

    droplet_creation_group.add_argument(
        '--ssh-key-id',
        metavar='ID',
        type=int,
        help='Greenhost-specific database ID of ssh key to deploy with')

    droplet_creation_group.add_argument(
        '--create-domain-records',
        action='store_true',
        help=('Creates DNS entries at Greenhost pointing the subdomain and '
              'domain to the cluster.'))

    droplet_creation_group.add_argument(
        '--subdomain',
        type=str,
        metavar='SUBDOMAIN',
        help=('Use a custom subdomain for the generated domain records. '
              'Defaults to no subdomain'))

    # Use this option to make sure that you won't run into troubles with
    # certificates. Let's Encrypt requires a CN for each certificate, and a
    # CN has a max length of 63 characters.
    droplet_creation_group.add_argument(
        '--truncate-subdomain',
        action='store_true',
        help=('Truncate subdomain so subdomain and domain are shorter than '
              '{} characters.'.format(MAX_DOMAIN_LENGTH)))

    droplet_creation_group.add_argument(
        '--disk-image-id',
        help=("Custom disk image ID. Use negative value for a custom template "
              "ID, use positive values for operating system IDs"))

    install_parser = subparsers.add_parser(
        'install',
        help=("Use this to run the ansible playbook that sets up your VPS to run "
              "Stackspin. The ansible-playbook process will run in the "
              "'{ansible_path}' directory, so do not use relative paths with these "
              "arguments.").format(ansible_path=ansible.ANSIBLE_PATH))
    install_parser.set_defaults(func=install)

    parser.add_argument(
        '--terminate-droplet',
        action='store_true',
        help=('Shutdown AND DELETE droplet identified by VPS identifier after '
              'finishing'))

    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help='Be more verbose')

    # Ansible related
    install_parser.add_argument(
        '--ansible-param',
        metavar=['PARAM[=VALUE]'],
        action='append',
        nargs=1,
        help='forward ansible parameters to the ansible-playbook call.')

    test_parser = subparsers.add_parser(
        'test',
        help=("Write test configuration and run tests on your cluster"))

    test_parser.set_defaults(func=test)

    test_parser.add_argument(
        '--observe',
        action='store_true',
        help=('Enables headful mode and runs script with 3000ms delay.'))
    test_parser.add_argument(
        '--apps',
        default='all',
        help=('Which apps to test (i.e. "nextcloud,wordpress". Defaults to'
              '"all".'))

    info_parser = subparsers.add_parser(
        'info',
        help=("Show information about a cluster"))
    info_parser.set_defaults(func=info)
    info_parser.add_argument(
        '--ip-address',
        action='store_true',
        help=('Only output the IP address of the machine')
    )

    secrets_parser = subparsers.add_parser(
        'secrets',
        help=("Show Stackspin cluster secrets"))
    secrets_parser.set_defaults(func=secrets)
    admin_credentials_parser = subparsers.add_parser(
        'admin-credentials',
        help=("Show Stackspin admin username and password"))
    admin_credentials_parser.set_defaults(func=admin_credentials)

    args = parser.parse_args()
    loglevel = logging.DEBUG if args.verbose else logging.INFO
    init_logging(log, loglevel)

    # Setup logging for greenhost_cloud module
    log_greenhost_cloud = logging.getLogger('greenhost_cloud')
    init_logging(log_greenhost_cloud, loglevel)

    log.debug("Parsed arguments: %s", str(args))

    clus = cluster.Cluster(args.cluster_name)

    if hasattr(args, 'func'):
        args.func(clus, args)

    if args.terminate_droplet:
        # In case none of the subparser's functions have been called, load data
        clus.load_data()
        greenhost_cloud.terminate_droplets_by_name('^{}$'.format(clus.hostname))

    if not hasattr(args, 'func') and not args.terminate_droplet:
        parser.print_help()
        sys.exit(1)

def info(clus, args):
    """
    Shows cluster information and then exits

    :param cluster.Cluster clus: cluster to show information about
    :param argparse.Namespace args: If the --ip-address argument is given, only
        prints the machine's IP address.
    """
    clus.load_data()
    clus.print_info(args)


def secrets(clus, args):
    """
    Dumps cluster secrets and then exits

    :param cluster.Cluster clus: cluster to show information about
    """
    clus.load_data()
    clus.dump_secrets()

def admin_credentials(clus, args):
    """
    Dumps admin username and password and then exits.

    :param cluster.Cluster clus: cluster to show information about
    """
    clus.load_data()
    clus.dump_admin_credentials()

def create(clus, args):  # pylint: disable=too-many-branches
    """
    Parses arguments for the 'create' subcommand

    :param cluster.Cluster clus: Cluster object to edit
    :param argparse.Namespace args: Command line arguments
    """
    clus = cluster.Cluster(args.cluster_name)

    if args.create_droplet:
        if not args.ssh_key_id:
            log.error('--ssh-key-id required when using --create-droplet')
            sys.exit(1)

    if args.subdomain:
        # Check if {subdomain}.{domain} is not too long to fit in CN
        if len(args.domain) > MAX_DOMAIN_LENGTH:
            log.error(('ERROR: domain argument is too long. Domain will not '
                       'fit in a CN, please make sure your subdomain + '
                       'domain + 1 do not exceed %d '
                       'characters'), MAX_DOMAIN_LENGTH)
            sys.exit(1)

        subdomain = args.subdomain
        if len(args.domain) + len(args.subdomain) + 1 > MAX_DOMAIN_LENGTH:
            if args.truncate_subdomain:
                half_required_length = floor((MAX_DOMAIN_LENGTH - len(args.domain) - 1)/2)
                subdomain = args.subdomain[:half_required_length]
                subdomain += args.subdomain[-half_required_length:]
                log.warning('Subdomain truncated to "%s"', subdomain)
            else:
                log.error(('ERROR: --subdomain argument is too long. Domain '
                           'will not fit in a CN, please make sure your '
                           'subdoman + domain + 1 do not exceed %d '
                           'characters'), MAX_DOMAIN_LENGTH)
                sys.exit(1)

        fqdn = '{subdomain}.{domain}'.format(
            subdomain=subdomain, domain=args.domain)
    else:
        fqdn = args.domain
        subdomain = None

    clus.domain = fqdn

    if args.disk_image_id:
        clus.disk_image_id = args.disk_image_id

    if args.create_droplet:
        clus.create_droplet(ssh_key_id=args.ssh_key_id, hostname=args.create_hostname)
        if args.verbose:
            greenhost_cloud.list_droplets()
        # Wait for ssh
        greenhost_cloud.wait_for_ssh(clus.ip_address)
    elif args.droplet_id:
        clus.set_info_by_droplet_id(args.droplet_id)
    elif args.ip_address:
        if args.create_hostname:
            create_hostname = args.create_hostname
        else:
            log.info('No --create-hostname provided, using cluster name "%s"',
                     args.cluster_name)
            create_hostname = args.cluster_name
        clus.set_info_by_ip_and_hostname(args.ip_address, create_hostname)
    elif args.droplet_hostname:
        clus.set_info_by_hostname(args.droplet_hostname)

    if args.docker_mirror_server:
        clus.docker_mirror_server = args.docker_mirror_server
    if args.docker_mirror_endpoint:
        clus.docker_mirror_endpoint = args.docker_mirror_endpoint
    if args.docker_mirror_username:
        clus.docker_mirror_username = args.docker_mirror_username
    if args.docker_mirror_password:
        clus.docker_mirror_password = args.docker_mirror_password

    # Write inventory.yml
    clus.write_cluster_files()

    if args.create_domain_records:
        create_domain_records(
            args.domain, clus.ip_address, subdomain=subdomain)
        if args.verbose:
            greenhost_cloud.list_domain_records(args.domain)


def install(clus, args):
    """
    Parses arguments for the 'install' subcommand

    :param cluster.Cluster clus: Cluster object to install Stackspin to
    :param argparse.Namespace args: Command line arguments
    """
    clus.load_data()

    ansible.run_ansible(
        clus,
        os.path.join(ansible.ANSIBLE_PATH, 'install-kubernetes.yml'),
        args.ansible_param)


def test(clus, args):
    """
    Runs taiko tests.

    :param cluster.Cluster clus: Cluster object to run tests on
    :param argparse.Namespace args: Command line arguments
    """

    taiko_path = os.path.join(os.path.dirname(__file__), '..', 'test', 'taiko')
    # Run from the taiko directory so taiko automatically loads all the
    # necessary files
    clus.load_data()
    command = ['taiko']
    if args.observe:
        command.append('--observe')
    command.append('apps.js')

    # Set env vars

    sso_username = clus.get_password_from_kubernetes(
        'stackspin-cluster-variables',
        'admin_email',
        'flux-system'
    )
    sso_password = clus.get_password_from_kubernetes(
        'stackspin-single-sign-on-variables',
        'userbackend_admin_password',
        'flux-system'
    )

    os.environ["SSO_USER_PW"] = sso_password
    os.environ["SSO_USERNAME"] = sso_username
    os.environ["TAIKO_TESTS"] = args.apps
    os.environ['DOMAIN'] = clus.domain

    log.info('Running taiko command %s', command)

    os.chdir(taiko_path)
    with Popen(command, stdout=PIPE, bufsize=1,
               universal_newlines=True) as taiko:
        for line in taiko.stdout:
            print(line, end='')

    return_code = (taiko.returncode)
    sys.exit(return_code)

def create_domain_records(domain, droplet_ip, subdomain=None):
    """
    Creates 2 domain records at Greenhost. An A record at subdomain.domain,
    pointing to droplet_ip and a CNAME record pointing *.subdomain.domain to
    the first record.

    :param str domain: the domain the cluster will be hosted on.
    :param str droplet_ip: The IP address the A record will point to.
    :param subdomain: Optional subdomain to host Stackspin on.
    :type subdomain: str or None
    """
    subdomain_arg = subdomain
    if subdomain is None:
        subdomain_arg = "@"

    domain_record = greenhost_cloud.create_domain_record(
        domain=domain, name=subdomain_arg,
        data=droplet_ip, record_type='A', update=True, ttl=60)
    log.info('Domain record: %s', domain_record)

    subdomain_arg = '*'
    if subdomain is not None:
        subdomain_arg += '.' + subdomain

    domain_record = greenhost_cloud.create_domain_record(
        domain=domain, name=subdomain_arg,
        data=subdomain, record_type='CNAME', update=True)
    log.info('Domain record: %s', domain_record)


def init_logging(logger, loglevel):
    """
    Configure logging.

    - debug and info go to stdout
    - warning and above go to stderr

    :param logger: Logger to initialise
    :param int loglevel: Log level from 0 to 50 (use logging.WARNING,
        logging.INFO, etc.)
    """
    logger.setLevel(loglevel)
    stdout = logging.StreamHandler(sys.stdout)
    stdout.setLevel(loglevel)
    stdout.addFilter(lambda record: record.levelno <= logging.INFO)
    logger.info('Set log level to %d', loglevel)

    stderr = logging.StreamHandler()
    stderr.setLevel(logging.WARNING)

    logger.addHandler(stdout)
    logger.addHandler(stderr)


if __name__ == '__main__':
    # Setup logging for this script
    log = logging.getLogger('stackspin')  # pylint: disable=invalid-name
    main()
